/*------------------------------------

	flash ROM writer firmware (1/3)
	
	2010.09.25 by Koichi Nishida

-------------------------------------*/

/*
This is a part of the firmware for the flash ROM writer by Nishida Radio.

Copyright (C) 2010 Koichi NISHIDA
email to Koichi NISHIDA: tulip-house@msf.biglobe.ne.jp

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <avr/io.h>
#include <avr/interrupt.h>

#define nop() __asm__ __volatile__ ("nop")
#define REVERSE(x) (((x&1)<<7)|((x&2)<<5)|((x&4)<<3)|((x&8)<<1)|((x&16)>>1)|((x&32)>>3)|((x&64)>>5)|((x&128)>>7))
#define FILE_SECTOR_NUM 512	// 512 sectors = 256 k bytes

// unions

union dwd {
	unsigned long l;
	unsigned char b[4];
};

union wd {
	unsigned short s;
	unsigned char b[2];
};

// C prototypes

// write a byte data to the SD card
void writeByteSD(unsigned char c);
// read data from the SD card
unsigned char readByteSD(void);
// issue SD card command without getting response
void cmd(unsigned char c, unsigned long adr);
// get command response from the SD card
unsigned char getResp(void);
// issue command 17 and get ready for reading
void cmd17(unsigned long adr);
// find a file extension
unsigned char findExt(void);
// get fat
void getFat(unsigned char ini);
// initialize SD card
unsigned char init(void);

// send a data to UART
void sendUART(unsigned char d);
// main function for the verify & write to the flash ROM
void verifywriteByteFlash(unsigned char isWrite);
// wait until data is written to the flash
void waitWrittenFlash(unsigned char d);

// reverse bits
unsigned char reverse(unsigned char c);
// read / write of Flash
unsigned char readByteFlash(void);
void writeByteFlash(unsigned char d);
// write header
void writeHeaderFlash(unsigned char d);

// assembler functions
void wait(void);

// SD card information
union dwd bpbAddr, rootAddr;
union wd fat, sectorsPerFat;

unsigned long fatAddr;							// the beginning of FAT
unsigned char sectorsPerCluster,spcShift;		// sectors per cluster	
unsigned long userAddr;							// the beginning of user data
unsigned short dir;
unsigned char slow;

// write a byte data to the SD card
void writeByteSD(unsigned char c)
{
	unsigned char d;
	for (d = 0b10000000; d; d >>= 1) {
		if (c&d) {
			PORTD = 0b00100000;
			if (slow) wait();
			PORTD = 0b00110000;
		} else {
			PORTD = 0b00000000;
			if (slow) wait();
			PORTD = 0b00010000;
		}
		if (slow) wait();
	}
	PORTD = 0b00000000;
}

// read data from the SD card
unsigned char readByteSD(void)
{
	unsigned char c = 0, i;

	PORTD = 0b00100000;
	if (slow) wait();
	for (i = 0; i != 8; i++) {
		PORTD = 0b00110000;
		if (slow) wait();
		c = ((c<<1) | (PINA&1));
		PORTD = 0b00100000;
		if (slow) wait();	
	}
	return c;
}

// issue a SD card command slowly without getting response
void cmd(unsigned char c, unsigned long adr)
{
	writeByteSD(0xff);
	writeByteSD(0x40+c);
	writeByteSD(adr>>24);
	writeByteSD((adr>>16)&0xff);
	writeByteSD((adr>>8)&0xff);
	writeByteSD(adr&0xff);
	writeByteSD(0x95);
	writeByteSD(0xff);
}

// issue a SD card command and wait normal response
void cmdResp(unsigned char c, unsigned long adr)
{
	unsigned char res;

	do {
		cmd(c, adr);
	} while (((res=getResp())!=0) && (res!=0xff));
}

// get a command response from the SD card
unsigned char getResp(void)
{
	unsigned char ch;

	do {
		ch = readByteSD();
	} while ((ch&0x80) != 0);
	return ch;
}

// issue command 17 and get ready for reading
void cmd17(unsigned long adr)
{
	unsigned char ch;

	cmd(17, adr);
	do {	
		ch = readByteSD();
	} while (ch != 0xfe);
}

// find a file extension
unsigned char findExt(void)
{
	unsigned short i;

	// find NIC extension
	for (i=0; i!=512; i++) {
		unsigned char ext[3], d;

		// check first char
		cmdResp(16, 1);
		cmd17(rootAddr.l+i*32);
		d = readByteSD();
		// readByteSD(); eadByteSD(); // discard CRC bytes
		if ((d==0x00)||(d==0x05)||(d==0x2e)||(d==0xe5)||(d=='.')) continue;
		cmdResp(16, 1);
		cmd17(rootAddr.l+i*32+11);
		d = readByteSD();
		// readByteSD(); readByteSD(); // discard CRC bytes
		if ((d&0x1e)|(d==0xf)) continue;
		// check extension
		cmdResp(16, 3);
		cmd17(rootAddr.l+i*32+8);
		ext[0] = readByteSD(); ext[1] = readByteSD(); ext[2] = readByteSD();
		// readByteSD(); readByteSD(); // discard CRC bytes
		if ((ext[0]=='R')&&(ext[1]=='O')&&(ext[2]=='M')) {
			dir = i;
			break;
		}	
	}
	return (i != 512);
}

// get fat
void getFat(unsigned char ini)
{
	unsigned long adr;
	
	adr = (ini?(rootAddr.l+dir*32+26):((unsigned long)fatAddr+(unsigned long)fat.s*2));
	cmdResp(16, 2);
	cmd17(adr);
	fat.b[0] = readByteSD();
	fat.b[1] = readByteSD();
	// readByteSD(); readByteSD(); // discard CRC bytes
	cmdResp(16,512);
}

// initialization called from check_eject
unsigned char init(void)
{
	unsigned char ch, i;
	char str[5];

	slow = 1;

	// initialize the SD card
	PORTD = 0b00000001;	
	for (i = 0; i != 200; i++) {
		PORTD = 0b00110001; 
		wait();
		PORTD = 0b00100001;
		wait();
	 }	// input 200 clock
 	PORTD = 0b00000000;
	
	cmd(0, 0);	// command 0
 	do {	
		ch = readByteSD();
	} while (ch != 0x01);

	PORTD = 0b00000001;
	while (1) {
		PORTD = 0b00000000;
		cmd(55, 0);	// command 55
		ch = getResp();
		if (ch == 0xff) return 0;
		if (ch & 0xfe) continue;
		// if (ch == 0x00) break;
		PORTD = 0b00000001;
		PORTD = 0b00000000;
		cmd(41, 0);	// command 41	
		if (!(ch=getResp())) break;
		if (ch == 0xff) return 0;
		PORTD = 0b00000001;
	}

	slow = 0;

	// BPB address
	cmdResp(16,5);
	cmd17(54);
	for (i=0; i<5; i++) str[i] = readByteSD();
	// readByteSD(); readByteSD(); // discard CRC
	if ((str[0]=='F')&&(str[1]=='A')/*&&(str[2]=='T')*/&&
		(str[3]=='1')&&(str[4]=='6')) {
		bpbAddr.l = 0;
	} else {
		cmdResp(16, 4);
		cmd17((unsigned long)0x1c6);
		bpbAddr.b[0] = readByteSD();
		bpbAddr.b[1] = readByteSD();
		bpbAddr.b[2] = readByteSD();
		bpbAddr.b[3] = readByteSD();
		bpbAddr.l *= 512;
		// readByteSD(); readByteSD(); // discard CRC bytes
	}

	// sectorsPerCluster and reservedSectors
	{
		union wd reservedSectors;	

		cmdResp(16, 3);
		cmd17(bpbAddr.l+0xd);		
		sectorsPerCluster = ch = readByteSD();

		spcShift=-1;
		while (ch) {
			spcShift++;
			ch>>=1;
		}

		reservedSectors.b[0] = readByteSD();
		reservedSectors.b[1] = readByteSD();
		// readByteSD(); readByteSD(); // discard CRC bytes	
		// sectorsPerCluster = 0x40 at 2GB, 0x10 at 512MB
		// reservedSectors.s = 2 at 2GB
		fatAddr = bpbAddr.l + (unsigned long)512*reservedSectors.s;
	}

	{
		// sectorsPerFat and rootAddr
		cmdResp(16, 2);
		cmd17(bpbAddr.l+0x16);
		sectorsPerFat.b[0] = readByteSD();
		sectorsPerFat.b[1] = readByteSD();
		// readByteSD(); readByteSD(); // discard CRC bytes		
		// sectorsPerFat =  at 512MB,  0xEF at 2GB
		rootAddr.l = fatAddr + ((unsigned long)sectorsPerFat.s*2*512);
		userAddr = rootAddr.l+(unsigned long)512*32;
	}

	// find "ROM" extension
	if (!findExt()) return 0;

	return 1;
}

void sendUART(unsigned char d)
{
	while (!( UCSRA & (1<<UDRE))) ;
	UDR = d;
}

int main(void)
{
	PORTB=0xff;
	DDRA = 0b00000000;
	DDRB = 0b00000000;
	DDRD = 0b00110011;

	// UART
	UBRRH = 0;
	UBRRL = 0;
	UCSRB = (1<<TXEN);
	UCSRC = (1<<USBS)|(3<<UCSZ0);

	while (1) {
		if (bit_is_clear(PIND,2))
			verifywriteByteFlash(1);
		else if (bit_is_clear(PIND, 3))
			verifywriteByteFlash(0);
	}
}

// reverse bits
unsigned char reverse(unsigned char c)
{
	unsigned char j, d=0;

	for (j=0; j!=8; j++) {
		d <<= 1;
		d |= (c&1);
		c >>= 1;
	}
	return d;
}

unsigned char readByteFlash(void)
{
	unsigned char d;

	sendUART(3);	// OE
	while (bit_is_set(PIND, 6));
	nop();
	d = PINB;	
	sendUART(4);	// OD
	while (bit_is_clear(PIND,6)) ;

	return d;
}

void writeByteFlash(unsigned char d)
{
	DDRB = 0b11111111;
	PORTB = d;
	sendUART(1);	// WE
	while (bit_is_set(PINA, 1)) ;
	sendUART(2);	// WD
	while (bit_is_clear(PINA, 1)) ;
	PORTB=0xff;
	DDRB = 0b00000000;
}

void writeHeaderFlash(unsigned char d)
{
	sendUART(0x55);
	writeByteFlash(REVERSE(0xaa));
	sendUART(0xaa);
	writeByteFlash(REVERSE(0x55));
	sendUART(0x55);
	writeByteFlash(reverse(d));
}

// wait until data is written to the flash
void waitWrittenFlash(unsigned char d)
{
	unsigned char c = reverse(d);

	while (readByteFlash() != c) ;
}

// main function for verify & write
void verifywriteByteFlash(unsigned char isWrite)
{
	unsigned short byte, cluster;
	unsigned char sector, match = 1;

	// initialize SD card
	if (init()) {
		sendUART(5);	// clear address
		sendUART(7);	// all LEDs off

		// LED (9: red or 8: green) on
		sendUART(8+isWrite);

		if (isWrite) {
			// chip erase
			writeHeaderFlash(0x80);	
			writeHeaderFlash(0x10);		
			waitWrittenFlash(0xff);
		}

		for (cluster = 0; cluster != (FILE_SECTOR_NUM>>spcShift); cluster++) {
			getFat(!cluster);
			for (sector = 0; sector != sectorsPerCluster; sector++) {
				cmd17(userAddr+(((unsigned long)(fat.s-2)<<spcShift)+sector)*512);
				for (byte = 0; byte != 512; byte++) {
					unsigned char d = readByteSD();
					
					if (isWrite) {
						writeHeaderFlash(0xa0);
						sendUART(0);	// address valid
						writeByteFlash(reverse(d));
						waitWrittenFlash(d);
					} else {
						unsigned char e;
						sendUART(0);	// address valid
						e = reverse(readByteFlash());
						if (e != d) match = 0;
					}
					sendUART(6);	// increment address	
				}
			}
		}
		sendUART(7);	// all LEDs off
		if (!match) sendUART(10);
	}
}